@@ -4,15 +4,16 @@
 
 use crate::actions::ActionSequence;
 use crate::capabilities::{
-    BrowserCapabilities, Capabilities, CapabilitiesMatching, LegacyNewSessionParameters,
-    SpecNewSessionParameters,
+    BrowserCapabilities, Capabilities, CapabilitiesMatching, SpecNewSessionParameters,
+};
+use crate::common::{
+    CredentialParameters, Date, FrameId, LocatorStrategy, ShadowRoot, WebElement, MAX_SAFE_INTEGER,
 };
-use crate::common::{Date, FrameId, LocatorStrategy, WebElement, MAX_SAFE_INTEGER};
 use crate::error::{ErrorStatus, WebDriverError, WebDriverResult};
 use crate::httpapi::{Route, VoidWebDriverExtensionRoute, WebDriverExtensionRoute};
 use crate::Parameters;
 use serde::de::{self, Deserialize, Deserializer};
-use serde_json::{self, Value};
+use serde_json::Value;
 
 #[derive(Debug, PartialEq)]
 pub enum WebDriverCommand<T: WebDriverExtensionCommand> {
@@ -41,7 +42,11 @@ pub enum WebDriverCommand<T: WebDriverExtensionCommand> {
     FindElements(LocatorParameters),
     FindElementElement(WebElement, LocatorParameters),
     FindElementElements(WebElement, LocatorParameters),
+    FindShadowRootElement(ShadowRoot, LocatorParameters),
+    FindShadowRootElements(ShadowRoot, LocatorParameters),
     GetActiveElement,
+    GetComputedLabel(WebElement),
+    GetComputedRole(WebElement),
     GetShadowRoot(WebElement),
     IsDisplayed(WebElement),
     IsSelected(WebElement),
@@ -73,8 +78,16 @@ pub enum WebDriverCommand<T: WebDriverExtensionCommand> {
     TakeScreenshot,
     TakeElementScreenshot(WebElement),
     Print(PrintParameters),
+    SetPermission(SetPermissionParameters),
     Status,
     Extension(T),
+    WebAuthnAddVirtualAuthenticator(AuthenticatorParameters),
+    WebAuthnRemoveVirtualAuthenticator,
+    WebAuthnAddCredential(CredentialParameters),
+    WebAuthnGetCredentials,
+    WebAuthnRemoveCredential,
+    WebAuthnRemoveAllCredentials,
+    WebAuthnSetUserVerified(UserVerificationParameters),
 }
 
 pub trait WebDriverExtensionCommand: Clone + Send {
@@ -167,6 +180,30 @@ impl<U: WebDriverExtensionRoute> WebDriverMessage<U> {
                 let element = WebElement(element_id.as_str().into());
                 WebDriverCommand::FindElementElements(element, serde_json::from_str(raw_body)?)
             }
+            Route::FindShadowRootElement => {
+                let shadow_id = try_opt!(
+                    params.get("shadowId"),
+                    ErrorStatus::InvalidArgument,
+                    "Missing shadowId parameter"
+                );
+                let shadow_root = ShadowRoot(shadow_id.as_str().into());
+                WebDriverCommand::FindShadowRootElement(
+                    shadow_root,
+                    serde_json::from_str(raw_body)?,
+                )
+            }
+            Route::FindShadowRootElements => {
+                let shadow_id = try_opt!(
+                    params.get("shadowId"),
+                    ErrorStatus::InvalidArgument,
+                    "Missing shadowId parameter"
+                );
+                let shadow_root = ShadowRoot(shadow_id.as_str().into());
+                WebDriverCommand::FindShadowRootElements(
+                    shadow_root,
+                    serde_json::from_str(raw_body)?,
+                )
+            }
             Route::GetActiveElement => WebDriverCommand::GetActiveElement,
             Route::GetShadowRoot => {
                 let element_id = try_opt!(
@@ -177,6 +214,24 @@ impl<U: WebDriverExtensionRoute> WebDriverMessage<U> {
                 let element = WebElement(element_id.as_str().into());
                 WebDriverCommand::GetShadowRoot(element)
             }
+            Route::GetComputedLabel => {
+                let element_id = try_opt!(
+                    params.get("elementId"),
+                    ErrorStatus::InvalidArgument,
+                    "Missing elementId parameter"
+                );
+                let element = WebElement(element_id.as_str().into());
+                WebDriverCommand::GetComputedLabel(element)
+            }
+            Route::GetComputedRole => {
+                let element_id = try_opt!(
+                    params.get("elementId"),
+                    ErrorStatus::InvalidArgument,
+                    "Missing elementId parameter"
+                );
+                let element = WebElement(element_id.as_str().into());
+                WebDriverCommand::GetComputedRole(element)
+            }
             Route::IsDisplayed => {
                 let element_id = try_opt!(
                     params.get("elementId"),
@@ -353,8 +408,26 @@ impl<U: WebDriverExtensionRoute> WebDriverMessage<U> {
                 WebDriverCommand::TakeElementScreenshot(element)
             }
             Route::Print => WebDriverCommand::Print(serde_json::from_str(raw_body)?),
+            Route::SetPermission => {
+                WebDriverCommand::SetPermission(serde_json::from_str(raw_body)?)
+            }
             Route::Status => WebDriverCommand::Status,
             Route::Extension(ref extension) => extension.command(params, &body_data)?,
+            Route::WebAuthnAddVirtualAuthenticator => {
+                WebDriverCommand::WebAuthnAddVirtualAuthenticator(serde_json::from_str(raw_body)?)
+            }
+            Route::WebAuthnRemoveVirtualAuthenticator => {
+                WebDriverCommand::WebAuthnRemoveVirtualAuthenticator
+            }
+            Route::WebAuthnAddCredential => {
+                WebDriverCommand::WebAuthnAddCredential(serde_json::from_str(raw_body)?)
+            }
+            Route::WebAuthnGetCredentials => WebDriverCommand::WebAuthnGetCredentials,
+            Route::WebAuthnRemoveCredential => WebDriverCommand::WebAuthnRemoveCredential,
+            Route::WebAuthnRemoveAllCredentials => WebDriverCommand::WebAuthnRemoveAllCredentials,
+            Route::WebAuthnSetUserVerified => {
+                WebDriverCommand::WebAuthnSetUserVerified(serde_json::from_str(raw_body)?)
+            }
         };
         Ok(WebDriverMessage::new(session_id, command))
     }
@@ -380,10 +453,11 @@ impl<U: WebDriverExtensionRoute> WebDriverMessage<U> {
                     } else {
                         let msg = format!("Failed to decode request as JSON: {}", body);
                         let stack = format!("Syntax error at :{}:{}", e.line(), e.column());
-                        Err(WebDriverError::new_with_stack(
+                        Err(WebDriverError::new_with_data(
                             ErrorStatus::InvalidArgument,
                             msg,
-                            stack,
+                            None,
+                            Some(stack),
                         ))
                     }
                 }
@@ -452,35 +526,28 @@ pub struct LocatorParameters {
     pub value: String,
 }
 
-/// Wrapper around the two supported variants of new session paramters.
-///
-/// The Spec variant is used for storing spec-compliant parameters whereas
-/// the legacy variant is used to store `desiredCapabilities`/`requiredCapabilities`
-/// parameters, and is intended to minimise breakage as we transition users to
-/// the spec design.
-#[derive(Debug, PartialEq)]
-pub enum NewSessionParameters {
-    Spec(SpecNewSessionParameters),
-    Legacy(LegacyNewSessionParameters),
+#[derive(Debug, PartialEq, Serialize)]
+pub struct NewSessionParameters {
+    capabilities: SpecNewSessionParameters,
 }
 
+// Manual deserialize implementation to error if capabilities is not an object
+// Without this the empty list test fails
 impl<'de> Deserialize<'de> for NewSessionParameters {
     fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
     where
         D: Deserializer<'de>,
     {
         let value = serde_json::Value::deserialize(deserializer)?;
-        if let Some(caps) = value.get("capabilities") {
-            if !caps.is_object() {
-                return Err(de::Error::custom("capabilities must be objects"));
-            }
-            let caps = SpecNewSessionParameters::deserialize(caps).map_err(de::Error::custom)?;
-            return Ok(NewSessionParameters::Spec(caps));
+        let caps = value
+            .get("capabilities")
+            .ok_or(de::Error::missing_field("capabilities"))?;
+        if !caps.is_object() {
+            return Err(de::Error::custom("capabilities must be objects"));
         }
-
-        warn!("You are using deprecated legacy session negotiation patterns (desiredCapabilities/requiredCapabilities), see https://developer.mozilla.org/en-US/docs/Web/WebDriver/Capabilities#Legacy");
-        let legacy = LegacyNewSessionParameters::deserialize(value).map_err(de::Error::custom)?;
-        Ok(NewSessionParameters::Legacy(legacy))
+        let capabilities =
+            SpecNewSessionParameters::deserialize(caps).map_err(de::Error::custom)?;
+        Ok(NewSessionParameters { capabilities })
     }
 }
 
@@ -489,10 +556,7 @@ impl CapabilitiesMatching for NewSessionParameters {
         &self,
         browser_capabilities: &mut T,
     ) -> WebDriverResult<Option<Capabilities>> {
-        match self {
-            NewSessionParameters::Spec(x) => x.match_browser(browser_capabilities),
-            NewSessionParameters::Legacy(x) => x.match_browser(browser_capabilities),
-        }
+        self.capabilities.match_browser(browser_capabilities)
     }
 }
 
@@ -502,6 +566,13 @@ pub struct NewWindowParameters {
     pub type_hint: Option<String>,
 }
 
+#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
+#[serde(untagged)]
+pub enum PrintPageRange {
+    Integer(u64),
+    Range(String),
+}
+
 #[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
 #[serde(default, rename_all = "camelCase")]
 pub struct PrintParameters {
@@ -511,7 +582,7 @@ pub struct PrintParameters {
     pub background: bool,
     pub page: PrintPage,
     pub margin: PrintMargins,
-    pub page_ranges: Vec<String>,
+    pub page_ranges: Vec<PrintPageRange>,
     pub shrink_to_fit: bool,
 }
 
@@ -529,19 +600,14 @@ impl Default for PrintParameters {
     }
 }
 
-#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
+#[derive(Clone, Debug, Default, PartialEq, Serialize, Deserialize)]
 #[serde(rename_all = "lowercase")]
 pub enum PrintOrientation {
     Landscape,
+    #[default]
     Portrait,
 }
 
-impl Default for PrintOrientation {
-    fn default() -> Self {
-        PrintOrientation::Portrait
-    }
-}
-
 #[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
 #[serde(default)]
 pub struct PrintPage {
@@ -584,6 +650,71 @@ impl Default for PrintMargins {
     }
 }
 
+#[derive(Debug, PartialEq, Serialize, Deserialize)]
+pub struct SetPermissionParameters {
+    pub descriptor: SetPermissionDescriptor,
+    pub state: SetPermissionState,
+}
+
+#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
+pub struct SetPermissionDescriptor {
+    pub name: String,
+}
+
+#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
+#[serde(rename_all = "lowercase")]
+pub enum SetPermissionState {
+    Denied,
+    Granted,
+    Prompt,
+}
+
+#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
+pub enum WebAuthnProtocol {
+    #[serde(rename = "ctap1/u2f")]
+    Ctap1U2f,
+    #[serde(rename = "ctap2")]
+    Ctap2,
+    #[serde(rename = "ctap2_1")]
+    Ctap2_1,
+}
+
+#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
+#[serde(rename_all = "kebab-case")]
+pub enum AuthenticatorTransport {
+    Usb,
+    Nfc,
+    Ble,
+    SmartCard,
+    Hybrid,
+    Internal,
+}
+
+fn default_as_true() -> bool {
+    true
+}
+
+#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
+#[serde(rename_all = "camelCase")]
+pub struct AuthenticatorParameters {
+    pub protocol: WebAuthnProtocol,
+    pub transport: AuthenticatorTransport,
+    #[serde(default)]
+    pub has_resident_key: bool,
+    #[serde(default)]
+    pub has_user_verification: bool,
+    #[serde(default = "default_as_true")]
+    pub is_user_consenting: bool,
+    #[serde(default)]
+    pub is_user_verified: bool,
+}
+
+#[derive(Clone, Debug, Serialize, Deserialize, PartialEq)]
+pub struct UserVerificationParameters {
+    #[serde(rename = "isUserVerified")]
+    pub is_user_verified: bool,
+}
+
 fn deserialize_to_positive_f64<'de, D>(deserializer: D) -> Result<f64, D::Error>
 where
     D: Deserializer<'de>,
@@ -613,7 +744,7 @@ pub struct SendKeysParameters {
 
 #[derive(Debug, PartialEq, Serialize, Deserialize)]
 pub struct SwitchToFrameParameters {
-    pub id: Option<FrameId>,
+    pub id: FrameId,
 }
 
 #[derive(Debug, PartialEq, Serialize, Deserialize)]
@@ -750,7 +881,7 @@ where
     let opt = Option::deserialize(deserializer)?.map(|value: f64| value as i64);
     let value = match opt {
         Some(n) => {
-            if n < i64::from(i32::min_value()) || n > i64::from(i32::max_value()) {
+            if n < i64::from(i32::MIN) || n > i64::from(i32::MAX) {
                 return Err(de::Error::custom(format!("'{}' is larger than i32", n)));
             }
             Some(n as i32)
@@ -768,7 +899,7 @@ where
     let opt = Option::deserialize(deserializer)?.map(|value: f64| value as i64);
     let value = match opt {
         Some(n) => {
-            if n < 0 || n > i64::from(i32::max_value()) {
+            if n < 0 || n > i64::from(i32::MAX) {
                 return Err(de::Error::custom(format!("'{}' is outside of i32", n)));
             }
             Some(n as i32)
@@ -1118,10 +1249,12 @@ mod tests {
             "alwaysMatch": {},
             "firstMatch": [{}],
         }});
-        let caps = NewSessionParameters::Spec(SpecNewSessionParameters {
-            alwaysMatch: Capabilities::new(),
-            firstMatch: vec![Capabilities::new()],
-        });
+        let caps = NewSessionParameters {
+            capabilities: SpecNewSessionParameters {
+                alwaysMatch: Capabilities::new(),
+                firstMatch: vec![Capabilities::new()],
+            },
+        };
 
         assert_de(&caps, json);
     }
@@ -1132,18 +1265,19 @@ mod tests {
         assert!(serde_json::from_value::<NewSessionParameters>(json).is_err());
     }
 
+    #[test]
+    fn test_json_new_session_parameters_capabilities_empty_list() {
+        let json = json!({ "capabilities": []});
+        assert!(serde_json::from_value::<NewSessionParameters>(json).is_err());
+    }
+
     #[test]
     fn test_json_new_session_parameters_legacy() {
         let json = json!({
             "desiredCapabilities": {},
             "requiredCapabilities": {},
         });
-        let caps = NewSessionParameters::Legacy(LegacyNewSessionParameters {
-            desired: Capabilities::new(),
-            required: Capabilities::new(),
-        });
-
-        assert_de(&caps, json);
+        assert!(serde_json::from_value::<NewSessionParameters>(json).is_err());
     }
 
     #[test]
@@ -1156,10 +1290,12 @@ mod tests {
             "desiredCapabilities": {},
             "requiredCapabilities": {},
         });
-        let caps = NewSessionParameters::Spec(SpecNewSessionParameters {
-            alwaysMatch: Capabilities::new(),
-            firstMatch: vec![Capabilities::new()],
-        });
+        let caps = NewSessionParameters {
+            capabilities: SpecNewSessionParameters {
+                alwaysMatch: Capabilities::new(),
+                firstMatch: vec![Capabilities::new()],
+            },
+        };
 
         assert_de(&caps, json);
     }
@@ -1173,10 +1309,12 @@ mod tests {
             },
             "foo": "bar",
         });
-        let caps = NewSessionParameters::Spec(SpecNewSessionParameters {
-            alwaysMatch: Capabilities::new(),
-            firstMatch: vec![Capabilities::new()],
-        });
+        let caps = NewSessionParameters {
+            capabilities: SpecNewSessionParameters {
+                alwaysMatch: Capabilities::new(),
+                firstMatch: vec![Capabilities::new()],
+            },
+        };
 
         assert_de(&caps, json);
     }
@@ -1264,6 +1402,92 @@ mod tests {
         assert!(serde_json::from_value::<PrintParameters>(json!({"scale": 3})).is_err());
     }
 
+    #[test]
+    fn test_json_permission() {
+        let params: SetPermissionParameters = SetPermissionParameters {
+            descriptor: SetPermissionDescriptor {
+                name: "push".into(),
+            },
+            state: SetPermissionState::Granted,
+        };
+        assert_de(
+            &params,
+            json!({"descriptor": {"name": "push"}, "state": "granted"}),
+        );
+    }
+
+    #[test]
+    fn test_json_permission_parameters_invalid() {
+        assert!(serde_json::from_value::<SetPermissionParameters>(json!({"test": 3})).is_err());
+    }
+
+    #[test]
+    fn test_json_permission_descriptor_invalid_type() {
+        assert!(serde_json::from_value::<SetPermissionParameters>(
+            json!({"descriptor": "test", "state": "granted"})
+        )
+        .is_err());
+    }
+
+    #[test]
+    fn test_json_permission_state_invalid_type() {
+        assert!(serde_json::from_value::<SetPermissionParameters>(
+            json!({"descriptor": {"name": "push"}, "state": 3})
+        )
+        .is_err());
+    }
+
+    #[test]
+    fn test_json_permission_state_invalid_value() {
+        assert!(serde_json::from_value::<SetPermissionParameters>(
+            json!({"descriptor": {"name": "push"}, "state": "invalid"})
+        )
+        .is_err());
+    }
+
+    #[test]
+    fn test_json_authenticator() {
+        let params = AuthenticatorParameters {
+            protocol: WebAuthnProtocol::Ctap1U2f,
+            transport: AuthenticatorTransport::Usb,
+            has_resident_key: false,
+            has_user_verification: false,
+            is_user_consenting: false,
+            is_user_verified: false,
+        };
+        assert_de(
+            &params,
+            json!({"protocol": "ctap1/u2f", "transport": "usb", "hasResidentKey": false, "hasUserVerification": false, "isUserConsenting": false, "isUserVerified": false}),
+        );
+    }
+
+    #[test]
+    fn test_json_credential() {
+        use base64::{engine::general_purpose::URL_SAFE, Engine};
+
+        let encoded_string = URL_SAFE.encode(b"hello internet~");
+        let params = CredentialParameters {
+            credential_id: r"c3VwZXIgcmVhZGVy".to_string(),
+            is_resident_credential: true,
+            rp_id: "valid.rpid".to_string(),
+            private_key: encoded_string.clone(),
+            user_handle: encoded_string.clone(),
+            sign_count: 0,
+        };
+        assert_de(
+            &params,
+            json!({"credentialId": r"c3VwZXIgcmVhZGVy", "isResidentCredential": true, "rpId": "valid.rpid", "privateKey": encoded_string, "userHandle": encoded_string, "signCount": 0}),
+        );
+    }
+
+    #[test]
+    fn test_json_user_verification() {
+        let params = UserVerificationParameters {
+            is_user_verified: false,
+        };
+        assert_de(&params, json!({"isUserVerified": false}));
+    }
+
     #[test]
     fn test_json_send_keys_parameters_with_value() {
         assert_de(
@@ -1294,23 +1518,36 @@ mod tests {
     }
 
     #[test]
-    fn test_json_switch_to_frame_parameters_with_value() {
+    fn test_json_switch_to_frame_parameters_with_number() {
         assert_de(
             &SwitchToFrameParameters {
-                id: Some(FrameId::Short(3)),
+                id: FrameId::Short(3),
             },
             json!({"id": 3}),
         );
     }
 
     #[test]
-    fn test_json_switch_to_frame_parameters_with_optional_null_field() {
-        assert_de(&SwitchToFrameParameters { id: None }, json!({ "id": null }));
+    fn test_json_switch_to_frame_parameters_with_null() {
+        assert_de(
+            &SwitchToFrameParameters { id: FrameId::Top },
+            json!({"id": null}),
+        );
     }
 
     #[test]
-    fn test_json_switch_to_frame_parameters_without_optional_null_field() {
-        assert_de(&SwitchToFrameParameters { id: None }, json!({}));
+    fn test_json_switch_to_frame_parameters_with_web_element() {
+        assert_de(
+            &SwitchToFrameParameters {
+                id: FrameId::Element(WebElement("foo".to_string())),
+            },
+            json!({"id": {"element-6066-11e4-a52e-4f735466cecf": "foo"}}),
+        );
+    }
+
+    #[test]
+    fn test_json_switch_to_frame_parameters_with_missing_id() {
+        assert!(serde_json::from_value::<SwitchToFrameParameters>(json!({})).is_err())
     }
 
     #[test]
@@ -1325,7 +1562,7 @@ mod tests {
             "foo": "bar",
         });
         let switch_to_frame = SwitchToFrameParameters {
-            id: Some(FrameId::Short(3)),
+            id: FrameId::Short(3),
         };
 
         assert_de(&switch_to_frame, json);
